! Module      : omcosflag_mod
!
! Authors     : Tim Harries (MSSL)
!             : Phil Williams (MSSL)
!             : Andrew King (MSSL)
!             : Chris Brindle (MSSL)
! Description : Applies shift and add to bad pixel map.
!
! Comments    : 
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
!
! History : Created  : January 98 - TJH
!           Modified : 27 April 98 - Removed twixel workaround.
!                                  - Removed datestamp call.
!                      20 March 99 - Spring clean of code.
!                      14 May 99 - Update shift and add process.
!                      
!
!                      24/08/2000 - AMK
!                      Initialised qualArraP
!
!                      24/08/2000 - AMK
!                      Changed incorrect initialisation of quality array.
!                      integer array had been initliased to real values
!
!                      Version 1.17.3: 14th June 2000
!                      Introduced new key word deAllocationError for warnings.
!
!                      Version 1.17.1: 9th June 2000
!                      Removed warning for bad pixel information used as test.
!                      We believe that the call to get` the bad pixel map works
!                      We do not know if it is correct

!                      Cleaned up memory allocation

!                      Removed forcing of one bad pixel - hardwired by JFR
!                      presumably for testing
!
!                      Version 1.15: 11th April 2000
!                      Modified 'zero drift' algorithm 		
!
!                      Version 1.14: 2nd March 2000
!                      Modified algorithm to cope with no drift file	
!	
!                      Version 1.13: 16th December 1999
!                      Updated test harness wrt new CCF for bad pixels
!                      Added new parameter timeSample	
!	
!                      Version 1.12: 2nd December 1999
!                      Added DISTRIBUTION (world) file	
!	
!                      Version 1.11: 8th June 1999
!                      Updated dependancies for V3.
!
!                      Version 1.10: 18th May 1999
!                      Removed CAL call to open CCF.
!                      Changed fatal datamode message.
!                      12.10.2001 - Program tidied up
!////////////////////////////////////////////////////////////////////
! 
! PATCHES AND FIXES AREA
!
! 1. 23.05.2000 - Removed forcing of bad pixels - believe that getBadPixelMap is working
!
!////////////////////////////////////////////////////////////////////
module omcosflag_mod

  use param
  use cal
  use dal
  use errorHandling
  use readHist_mod
  use readModes_mod

  use omallocation_mod
use omutility_mod
  implicit none

  private

  public omcosflag

contains

subroutine omcosflag
    
    type(DataSetT) :: inDataSetH                                      ! Input dataset handle.
    type(DataSetT) :: thxDataSetH                                     ! Tracking History Data Auxiliary dataset handle.
    type(TableT) :: inTableH                                          ! Input table handle.
    type(TableT) :: thxTableH                                         ! Tracking History Data Auxiliary table handle.
    type(ColumnT), dimension(4) :: inColumnH                          ! Input image column handle.
    type(ColumnT), dimension(3) :: thxColumnH                         ! Tracking History Data Auxiliary column handle.
    type(ArrayT) :: qualArrayH                                        ! Quality Array handle.
    type(ArrayT) :: imageArrayH                                       ! Image array handle
    real(kind=single), pointer :: imageArrayP(:, :)        ! Pointer to image array

    type trackingFrame
       real(single) :: dX                                        ! Delta x.
       real(single) :: dY                                        ! Delta y.
       real(single) :: quality                                   ! Quality.
    end type trackingFrame
    type(trackingFrame), pointer:: tf(:)=>Null()              ! Tracking frame array.


    character(8) :: dataMode                                      ! Input file data mode.
    character(255) :: set                                         ! OM OSW Image filename.
    character(255) :: thxSet                                      ! OM Tracking History Data Auxiliary filename.

    integer(int8), pointer :: qualArrayP(:, :)=>Null()        ! Quality array pointer.

    integer(int16), pointer :: badPixMapP(:, :)=>Null()       ! Bad pixel map pointer.
    integer(int16), pointer :: windowedbadPixMap(:, :)=>Null() ! Bad pixel map.
    integer(int16), pointer :: binnedbadPixMap(:, :)=>Null() ! Bad pixel map.
    integer(int16) :: nFrames                                    ! Number of tracking frames. 

    integer(int32) :: sampleFactor                               ! Sampling factor.
    integer(int32) :: timeFactor                                 ! Time sampling factor
    integer(int32), pointer :: windowX0P(:)=>Null()   ! WindowX0 pointer.
    integer(int32), pointer :: windowY0P(:)=>Null()   ! WindowY0 pointer.
    integer(int32), pointer :: windowDXP(:)=>Null()   ! WindowDX pointer.
    integer(int32), pointer :: windowDYP(:)=>Null()   ! WindowDY pointer.
    integer(int32), pointer :: dXP(:)=>Null()         ! DX pointer.
    integer(int32), pointer :: dYP(:)=>Null()         ! DY pointer.
    integer(int32), pointer :: qualityP(:)=>Null()    ! Quality pointer.
    integer(int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(int32) :: binAx2                                     ! DPU y-axis binning factors.
    integer(int32) :: windowX0                                   ! WindowX0.
    integer(int32) :: windowY0                                   ! WindowY0.
    integer(int32) :: windowDX                                   ! WindowDX.
    integer(int32) :: windowDY                                   ! WindowDY.
    integer(int32) :: allocErr                                   ! Allocate error status.
    integer(int32) :: maxDX                                      ! Maximum drift in x (centroided pixels).
    integer(int32) :: maxDY                                      ! Maximum drift in y (centroided pixels).
    integer(int32) :: cpxMin                                     ! Minimum x value for window (centroided pixels).
    integer(int32) :: cpxMax                                     ! Maximum x value for window (centroided pixels).
    integer(int32) :: cpyMin                                     ! Minimum y value for window (centroided pixels).
    integer(int32) :: cpyMax                                     ! Maximum y value for window (centroided pixels).
    integer(int32) :: i                                          ! Count.
    integer(int32) :: j                                          ! Count.
    integer(int32) :: oMaxDX                                     ! Maximum drift in x (oversampled pixels).
    integer(int32) :: oMaxDY                                     ! Maximum drift in y (oversampled pixels).
    integer(int32) :: opxMin                                     ! Minimum x value for window (oversampled pixels).
    integer(int32) :: opxMax                                     ! Maximum x value for window (oversampled pixels).
    integer(int32) :: opyMin                                     ! Minimum y value for window (oversampled pixels).
    integer(int32) :: opyMax                                     ! Maximum y value for window (oversampled pixels).
    integer(int32) :: shiftX                                     ! Shift in x.
    integer(int32) :: shiftY                                     ! Shift in y.
    integer(int32) :: x                                          ! X.
    integer(int32) :: y                                          ! Y.
    integer(int32) :: iShiftX                                    ! Incremental shift in x.
    integer(int32) :: iShiftY                                    ! Incremental shift in y.
    integer(int32) :: iXMin                                      ! Incremental minimum x value for window.
    integer(int32) :: iXMax                                      ! Incremental maximum x value for window.
    integer(int32) :: iYMin                                      ! Incremental minimum y value for window.
    integer(int32) :: iYMax                                      ! Incremental maximum y value for window.
    integer(int32) :: xConstant, yConstant
    integer(int32) :: qualArrayDim(2)                 ! Quality array dimension.
    integer(int32) :: binX                                       ! Bin x.
    integer(int32) :: binY                                       ! Bin y.
    integer(int32) :: iPix
    integer(int32), pointer :: yIndex(:)=>Null()  ! Stores j array-element indices               
    real(single) :: rShiftX                                      ! Relative shift in x.
    real(single) :: rShiftY                                      ! Relative shift in y.
    real(single) :: missingPixelValue
    logical(bool) :: binBPE                                      ! BPE binning flag.
    logical(bool), pointer :: badPixMap(:, :)=>Null() !bad  pixel map.
    logical(bool), pointer :: lBadPixMap(:, :)=>Null()    ! Localised bad pixel map.
    logical(bool), pointer :: oBadPixMap(:, :)=>Null()    ! Oversampled bad pixel map.
    logical(bool), pointer :: oQualArray(:, :)=>Null()    ! Oversampled quality array.
    logical(bool), pointer :: qualArray(:, :)=>Null()     ! Quality array.
    integer iMax, jMax
    integer            :: istat    ! Status flag
    integer :: binnedBadPixMapDim(2)   

    !**********************************
    ! Read in parameters.
    !**********************************
    sampleFactor = intParameter("samplefactor")
    timeFactor = intParameter("timefactor")
    set = stringParameter("set")


    thxSet = stringParameter("thxset")

    ! Read in the OM OSW Image file.

    inDataSetH = dataSet(set, modify)

    ! Determine file type from datamode in primary header.

    dataMode = stringAttribute(inDataSetH, "DATAMODE")


    !*************************************
    ! Check data mode okay
    !*************************************   
    call CheckValidDataMode(dataMode, "IMAGING", set)
     
    !*************************************************************
    ! Read in the science window keywords from the primary header.
    !*************************************************************
    binBPE = booleanAttribute(inDataSetH, "BINBPE")
    binAx1 = int32Attribute(inDataSetH, "BINAX1")
    binAx2 = int32Attribute(inDataSetH, "BINAX2")


    ! Get a handle on the modes binary table extension.

    inTableH = table(inDataSetH, "MODES")

    ! Get handles and pointers on the modes binary table extension columns and extract the window data.

    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)


    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)

    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)

    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)

    !**********************************************************
    ! 25.5.2001 - Need the missing-pixel value for the badpixelmap
    !**********************************************************
    missingPixelValue = real32Attribute(inDataSetH, "MISSPIX")
 

 
    !*************************************************
    ! Need to examine the image pixels for missing 
    ! pixels
    !************************************************* 
    imageArrayH = array(inDataSetH, "PRIMARY", read) 
    imageArrayP => real32Array2Data(imageArrayH)
 
    ! Release handles on data.
 
    do i = 1,  4
        call release(inColumnH(i))
    end do 
    call release(inTableH)

    !*****************************************************
    ! Read in the OM Tracking History Data Auxiliary file.
    !*****************************************************
    thxDataSetH = dataSet(thxSet, read)





    call CAL_setState(OM, 1, 0)

    !*******************************************************************
    ! Get the bad-pixel map
    !******************************************************************* 
    call CAL_getBadPixelMap(badPixMapP)   


    if (hasBlock(thxDataSetH, "ZERODRIFT")) then



       ! Issue warning

       write (errstr, *) "Missing OM Tracking History Data Auxiliary File. Using ZERODRIFT option"      
       call message(AppMsg, Verbose)

     
       !******************************
       ! Calculate the image binning.
       !******************************
       binX = 2 ** binAx1
       binY = 2 ** binAx2

       if (binBPE) then
          binX = binX * 2
          binY = binY * 2
       end if

       qualArrayDim(1) = windowDX / binX   ! The x dimensions of the quality array
       qualArrayDim(2) = windowDY / binY   ! The y dimensions of the quality array


       allocate (windowedBadPixMap(0 : qualArrayDim(1) - 1, 0 : &
                                       qualArrayDim(2) - 1), stat=allocErr)
       call CheckAllocation(allocErr, "windowedBadPixelMap")



        !********************************************  
        ! Bin the fullbad pixmap image
        !*********************************************

        binnedBadPixMap => BinBadPixMap(badPixMapP, binX, binY)
        binnedBadPixMapDim(1) = size(binnedBadPixMap, 1)
        binnedBadPixMapDim(2) = size(binnedBadPixMap, 2)


       !********************************************
       ! window the full image to the input image.
       !********************************************
       yIndex => AllocateInt32(0,  qualArrayDim(2) - 1, "yIndex")
       
       
       forall (j = 0 : (qualArrayDim(2) - 1)) yIndex(j) =  min(binnedBadPixMapDim(2), windowY0 / binY + j) 

       do i = 0, qualArrayDim(1) - 1
          iPix = min(binnedBadPixMapDim(1), windowX0 / binX + i) 
          do j = 0, qualArrayDim(2) - 1
              windowedBadPixMap(i, j) = binnedBadPixMap(iPix, yIndex(j))
  
          end do
       end do



       deallocate(yIndex, stat=allocErr)
       call CheckDeallocation(allocErr, "yIndex")
       nullify(yIndex)

       !*********************************************************************
       ! Add the quality array
       !*******************************************************************
       qualArrayH = addArray(inDataSetH, "QUALITY", Integer8, qualArrayDim)
       qualArrayP => int8Array2Data(qualArrayH)
   
       

       ! Convert to a boolean bad pixel map.

       !
       !--------------------------------------------------------------------------------------------
       !
       ! AMK - 24/08/200
       !
       ! Changed initialisation below to correct integer valuews and NOT reals
       !
       !      qualArrayP = 0.0
       !      where (windowedBadPixMap /= 0) qualArrayP = 1.0

       qualArrayP = 0


       !**************************************************
       ! Check the image for missing pixels or negative
       ! values- if any set corresponding quality-map pixel
       ! to 1
       !**************************************************
       call PixelChecks(imageArrayP, qualArrayP, missingPixelValue)


       !***********************************************************
       ! Check for bad pixel columns
       !***********************************************************
     
       call BadColumnChecks(imageArrayP, qualArrayP, binx)

     
       where (windowedBadPixMap .ne. 0) qualArrayP = 1
 
    

       call release(qualArrayH)

    else   ! NONE ZERO-DRIFT BLOCK
       ! Determine file type from datamode in primary header.



       dataMode = stringAttribute(thxDataSetH, "DATAMODE")

       !********************************
       ! Check data mode okay
       !******************************** 
       call CheckValidDataMode(dataMode, "TRACKING", thxSet)
     
       ! Get a handle on the tracking history binary table extension.

       thxTableH = table(thxDataSetH, "OMTHX1")

       ! Get the number of frames in the tracking history binary table extension.

       nFrames = numberOfRows(thxTableH)

       

       ! Allocate space for the tracking frames.

       allocate (tf(0 : nFrames), stat=allocErr)
       call CheckAllocation(allocErr, "tf")

  

       ! Get handles and pointers on the tracking history binary table extension columns and extract the tracking data.

       thxColumnH(1) = column(thxTableH, "DX", read)
       dXP => int32Data(thxColumnH(1))
       tf(0)%dX = 0.0
       tf(1 : nFrames)%dX = real(dXP / 1000.0)

       thxColumnH(2) = column(thxTableH, "DY", read)
       dYP => int32Data(thxColumnH(2))
       tf(0)%dY = 0.0
       tf(1 : nFrames)%dY = real(dYP / 1000.0)

       thxColumnH(3) = column(thxTableH, "QUALITY", read)
       qualityP => int32Data(thxColumnH(3))
       tf(0)%quality = 0.0
       tf(1 : nFrames)%quality = real(qualityP / 1000000.0)

       ! Release handles on data.

       do i = 1,  3 
           call release(thxColumnH(i))
       end do

       call release(thxTableH)
       call release(thxDataSetH)

   

      ! Allocate space for the boolean bad pixel map.

       badPixMap => Allocate2DBoolean(0, 2047, 0, 2047, "badPixMap")
       


       ! Convert to a boolean bad pixel map.

       badPixMap = .false.
       where (badPixMapP /= 0) badPixMap = .true.
  

       ! Calculate the maximum drift in x and y.
       ! We add 0.5 to the drifts so that positions are calculated with respect to the centre of the
       ! pixel rather than the bottom left hand corner.

       maxDX = 0
       maxDY = 0
       do i = 1, nFrames
          maxDX = max(maxDX, ceiling(0.5 + abs(tf(i)%dX)))
          maxDY = max(maxDY, ceiling(0.5 + abs(tf(i)%dY)))
       end do

       ! Allocate space for the loaclised bad pixel map.

       cpxMin = windowX0 - maxDX
       cpxMax = windowX0 + windowDX - 1 + maxDX
       cpyMin = windowY0 - maxDY
       cpyMax = windowY0 + windowDY - 1 + maxDY
   
       lBadPixMap => Allocate2DBoolean(cpxMin, cpxMax, cpyMin, cpyMax, "lBadPixMap")
        



       ! Fill localised bad pixel map.

       lBadPixMap = .false.
       do i = max(0, cpxMin) , min(2047, cpxMax)
           do j = max(0, cpyMin), min(2047, cpyMax)
                lBadPixMap(i, j) = badPixMap(i, j)
           end do 
       end do

       ! Free up memory.

       deallocate(badPixMap, stat=allocErr)
       call CheckDeallocation(allocErr, "badPixMap")


       ! Allocate space for the oversampled bad pixel map.

       oMaxDX = maxDX * sampleFactor
       oMaxDY = maxDY * sampleFactor
       opxMin = 0 - oMaxDX
       opxMax = ((cpxMax - cpxMin - (2 * maxDX)) * sampleFactor) + oMaxDX
       opyMin = 0 - oMaxDY
       opyMax = ((cpyMax - cpyMin - (2 * maxDY)) * sampleFactor) + oMaxDY
       oBadPixMap => Allocate2DBoolean(opxMin, opxMax, opyMin, opyMax, "oBadPixMap")
      
 

       !**********************************
       ! Oversample the bad pixel map.
       !*********************************
       oBadPixMap = .false.

       yIndex => AllocateInt32(opyMin, opyMax, "yIndex")
       
       forall (j = opyMin : opyMax) yIndex(j) =  floor((j + oMaxDY) / real(sampleFactor)) + cpyMin

       do i = opxMin, opxMax           
          shiftX = floor((i + oMaxDX) / real(sampleFactor))
          x = cpxMin + shiftX
          do j = opyMin, opyMax      
              oBadPixMap(i, j) = lBadPixMap(x, yIndex(j))

          end do
       end do
      


       ! Free up memory.

       deallocate(lBadPixMap, stat=allocErr)
       call CheckDeallocation(allocErr, "lBadPixMap")


       ! Allocate space for the oversampled quality array.

       opxMin = 0
       opxMax = (cpxMax - cpxMin - (2 * maxDX)) * sampleFactor
       opyMin = 0
       opyMax = (cpyMax - cpyMin - (2 * maxDY)) * sampleFactor

       oQualArray => Allocate2DBoolean(opxMin, opxMax, opyMin, opyMax, "oQualArray")
   
       !****************************************************************************************** 
       ! Loop over the tracking frames and construct an oversampled tracking shifted quality array.
       ! When we apply the shifts, we must centre the star in the pixel, so we add 0.5 to x
       ! and y or in this case to dx and dy.
       !******************************************************************************************

       oQualArray = .false.
       do i = 1, nFrames
          rShiftX = (tf(i)%dx - tf(i - 1)%dx) * sampleFactor
          rShiftY = (tf(i)%dy - tf(i - 1)%dy) * sampleFactor
          xConstant =  (0.5 + tf(i - 1)%dx) * sampleFactor
          yConstant =  (0.5 + tf(i - 1)%dy) * sampleFactor
          do j = 0, timeFactor
    
             iShiftX = nint(xConstant + (rShiftX * (j / real(timeFactor))))
             iShiftY = nint(yConstant + (rShiftY * (j / real(timeFactor))))
             iXMin = opxMin - iShiftX
             iXMax = opxMax - iShiftX
             iYMin = opyMin - iShiftY
             iYMax = opyMax - iShiftY
             oQualArray = oQualArray .or. oBadPixMap(iXMin : iXMax, iYMin : iYMax)
          end do
       end do

       ! Free up memory.

       deallocate(oBadPixMap, stat=allocErr)
       call CheckDeallocation(allocErr, "oBadPixMap")

       ! Allocate space for the quality array.

       cpxMin = windowX0
       cpxMax = windowX0 + windowDX - 1
       cpyMin = windowY0
       cpyMax = windowY0 + windowDY - 1

       qualArray => Allocate2DBoolean(cpxMin, cpxMax, cpyMin, cpyMax, "qualArray")
   
 
       !**************************
       ! Fill the quality array.
       !**************************

       qualArray = .false.
       opxMin = 0
       opxMax = (cpxMax - cpxMin) * sampleFactor
       opyMin = 0
       opyMax = (cpyMax - cpyMin) * sampleFactor

       forall (j = opyMin : opyMax) yIndex(j) = cpyMin + floor(j / real(sampleFactor)) 

       do i = opxMin,  opxMax
          x = cpxMin + floor(i / real(sampleFactor))
          do j = opyMin, opyMax
              qualArray(x, yIndex(j)) = oQualArray(i, j)
              
          end do
       end do

       deallocate(yIndex, oQualArray, stat=allocErr)
       call CheckDeallocation(allocErr, "yIndex")
       nullify(yIndex)

  
       ! Calculate the image binning.

       binX = 2**binAx1
       binY = 2**binAx2

       if (binBPE) then
          binX = binX * 2
          binY = binY * 2
       end if

       qualArrayDim = shape(qualArray)

       qualArrayDim(1) = qualArrayDim(1) / binX
       qualArrayDim(2) = qualArrayDim(2) / binY

       ! Add the quality extension to the image file.

       qualArrayH = addArray(inDataSetH, "QUALITY", Integer8, qualArrayDim)
       qualArrayP => int8Array2Data(qualArrayH)

       !********************************
       ! Bin the quality array.
       !********************************
       qualArrayP = 0
   
       yIndex => AllocateInt32(cpyMin, cpyMax, "yIndex")

       iMax =  qualArrayDim(1) - 1
       jMax =  qualArrayDim(2) - 1

       do j = cpyMin, cpyMax 
           shiftY = j - cpyMin
           y = floor(shiftY / real(binY))
           yIndex(j) = min(y, jMax)    
       end do
    
       do i = cpxMin, cpxMax
          shiftX = i - cpxMin
          x = floor(shiftX / real(binX))
          x = min(x, iMax)       ! Added 12.7.2000 to prevent array bound error
          do j = cpyMin, cpyMax
    
        
  
             if (qualArray(i, j)) qualArrayP(x, yIndex(j)) = 1
             
          end do
       end do

       !************************************************************
       !
       !**********************************************************
       call PixelChecks(imageArrayP, qualArrayP, missingPixelValue)

       !***********************************************************
       ! Check for bad pixel columns
       !***********************************************************
     
       call BadColumnChecks(imageArrayP, qualArrayP, binx)


       deallocate(yIndex, stat=allocErr)
       nullify(yIndex)    

       ! Release handles on data.

       call release(qualArrayH)

    endif

    ! Release the cal memory.
    call cal_releaseMemory(badPixMapP)


    call release(imageArrayH)
    call release(inDataSetH)

   

    if(associated(binnedBadPixMap)) deallocate(binnedBadPixMap, stat = istat)
    if(associated(windowedBadPixMap)) deallocate(windowedBadPixMap,stat=istat)
    if(associated(tf)) deallocate(tf, stat = istat)
    if(associated(badPixMap)) deallocate(badPixMap,stat = istat)
    if(associated(lBadPixMap)) deallocate(lBadPixMap,stat = istat)
    if(associated(oQualArray)) deallocate(oQualArray,stat = istat)
    if(associated(qualArray)) deallocate(qualArray,stat = istat)
    call CheckDeallocation(istat, "qualArray")


end subroutine omcosflag

!**********************************
! Subroutine CheckValidDataMode
!**********************************
subroutine CheckValidDataMode(dataMode, validMode, fileName)
    character(*), intent(in) :: dataMode
    character(*), intent(in) :: validMode
    character(*), intent(in) :: fileName


    if (trim(dataMode) .ne. trim(validMode)) then
 
        write (errstr, *) "Invalid datamode ", dataMode, " in set ", fileName
        call fatal("badFileMode")
    end if

end subroutine CheckValidDataMode

!*****************************************************
! function BinBadPixMap
!*****************************************************
function BinBadPixMap(badPixMapP, binX, binY) result(binnedBadPixMapP)
    integer(int16), pointer :: badPixMapP(:, :) ! Bad pixel map pointer.

    integer(int32), intent(in) :: binX, binY                                       ! Bin x.
    

    integer(int16), pointer :: binnedBadPixMapP(:, :) ! binned bad pixel map pointer.

    integer(int32), pointer :: yIndex(:)=>Null()  ! Stores j array-element indices               
    
    integer(int32) :: i, j, iPix, jPix
  
    
    binnedBadPixMapP => Allocate2DInt16(0, 2047 / binX, 0, 2047 / binY, "binnedBadPixMap")
                              
    
    if (binX .eq. 1 .and. binY .eq. 1) then
        binnedBadPixMapP = badPixMapP
        return
    end if     
    !***********************  
    ! Bin the full image
    !***********************

    yIndex => AllocateInt32(0 , 2047, "yIndex")
       

    forall (j = 0 : 2047)  yIndex(j) =  j / binY


    do i = 0, 2047

        iPix = i / binX

        do j = 0, 2047

            jPix = yIndex(j)
            binnedBadPixMapP(iPix, jPix) = binnedBadPixMapP(iPix, jPix) + badPixMapP(i, j)
         end do
    end do

   deallocate(yIndex)
       


end function BinBadPixMap






!******************************************************
!
!
! Check the image for missing pixels or negative
! values- if any set corresponding quality-map pixel
! to 1
!**************************************************
subroutine PixelChecks(imageArrayP, qualArrayP, missingPixelValue)
    real(single), pointer :: imageArrayP(:, :)        ! Pointer to image array
    integer(int8), pointer :: qualArrayP(:, :)        ! Quality array pointer.
    real(single), intent(in) :: missingPixelValue
    integer :: i, j

 
    do i = 0, size(imageArrayP, 1) - 1
              
        do j = 0, size(imageArrayP, 2) - 1
      
            if (abs(imageArrayP(i, j) - missingPixelValue) .lt. 0.5) then
                qualArrayP(i, j) = 1
                imageArrayP(i, j) = 0.0 
            else
                if (imageArrayP(i, j) .lt. 0.0) then

                    qualArrayP(i, j) = 1
                    imageArrayP(i, j) = 0.0 
                end if
            end if
        end do
    end do
end subroutine PixelChecks       






!***********************************************************
! Subroutine BadColumnChecks
! Check for bad pixel columns
!***********************************************************
     
subroutine BadColumnChecks(image, qualImage, binx)    
    real(single), pointer :: image(:, :)        ! Pointer to image array
    integer(int8), pointer :: qualImage(:, :)        ! Quality array pointer.
    integer(int32), intent(in) :: binx
    integer(int16) :: i, j
    integer(int16) :: nx, ny
    integer :: count

    nx = ubound(image, 1)
    ny = ubound(image, 2) 
    count = 0
    do i = 0, nx
              
        if (all(image(i, :) .lt. 0.0001)) then
            qualImage(i, :) = 1
            count = count + 1
            cycle
        end if
        if (i == 0 .or. i == nx) cycle
        do j = 1, ny - 1
            if (qualImage(i, j) > 0) cycle          
            call checkforHotPixels(nx, ny, image, qualImage, i, j, binx)
          
        end do



    end do

    if (count .eq. 0) return

    if (count .eq. 1) then
        write (errstr, *) "1 bad column was found"      
    else
        write (errstr, '(i4, a)') count,  " bad columns were found"      
        
    end if
    call message(AppMsg, Verbose)


end subroutine BadColumnChecks










!**************************************************
!
!
!*************************************************
subroutine checkforHotPixels(nx1, ny1, imagePtr, qualImage, i0, j0, binx)
    integer(int16), intent(in) :: nx1, ny1
    real(single), pointer :: imagePtr(:, :)
    integer(int8), pointer :: qualImage(:, :)        ! Quality array pointer.



    integer(int16), intent(in) :: i0, j0
    integer, intent(in) :: binx
    integer(int16) :: dx, il, iu, jl, ju
    integer :: i, j
    real(single) :: f0, f, f1 


    f0 = imagePtr(i0, j0)
    if (f0 < 30.0) return


    dx = 12 / binx


    il = i0 - dx
    jl = j0 - dx


    il = max(0_int16, il)
    jl = max(0_int16, jl)


    iu = i0 + dx
    ju = j0 + dx

    iu = min(nx1, iu)
    ju = min(ny1, ju)
    
       
    f1 = 0.0
    do i = il, iu

        do j = jl, ju
            if (i == i0 .and. j == j0) cycle 
	    f = 20.0 * max(1.0, imagePtr(i, j))
                       
            if (f > f0) return 
            f1 = max(f1, f)
        end do
    end do



 
       
    qualImage(i0, j0) = 1


!write(*, *) "omcosflag pixel(", i0, ", ", j0, ") set to 1- f0=", f0, " f1=", f1   


 






end subroutine checkforHotPixels

end module omcosflag_mod













